"""Definition for middleware that precess next step handlers logic."""

import contextlib
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union
from uuid import UUID

from loguru import logger
from pydantic import BaseConfig, BaseModel

from botx import Bot, Collector, concurrency, converters, exceptions
from botx.collecting.handlers.handler import Handler
from botx.collecting.handlers.name_generators import get_name_from_callable
from botx.dependencies.models import Depends
from botx.middlewares.base import BaseMiddleware
from botx.models.messages.message import Message
from botx.typing import Executor


class NextStepHandlerState(BaseModel):
    """Information about next step handler."""

    class Config(BaseConfig):
        arbitrary_types_allowed = True

    #: name of handler that should be called.
    name: str

    #: arguments that should be set on message for handler.
    arguments: Dict[str, Any]


class NextStepMiddleware(BaseMiddleware):
    """
    Naive next step handlers middleware. May be useful in simple apps or as base.

    Important:
        This middleware should be the last included into bot, since it will break
        execution if right handler will be found.
    """

    def __init__(  # noqa: WPS211
        self,
        executor: Executor,
        bot: Bot,
        functions: Union[Dict[str, Callable], Sequence[Callable]],
        break_handler: Optional[Union[Handler, str, Callable]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        dependency_overrides_provider: Any = None,
    ) -> None:
        """Init middleware with required query_params.

        Arguments:
            executor: next callable that should be executed.
            bot: bot that will store ns state.
            functions: dict of functions and their names that will be used as next step
                handlers or set of sequence of functions that will be registered by
                their names.
            break_handler: handler instance or name of handler that will break next step
                handlers chain.
            dependencies: background dependencies that should be applied to handlers.
            dependency_overrides_provider: object that will override dependencies for
                this handler.
        """
        super().__init__(executor)

        dependencies = converters.optional_sequence_to_list(
            bot.collector.dependencies,
        ) + converters.optional_sequence_to_list(dependencies)
        dep_override = (
            dependency_overrides_provider or bot.collector.dependency_overrides_provider
        )
        bot.state.ns_collector = Collector(
            dependencies=dependencies, dependency_overrides_provider=dep_override,
        )
        bot.state.ns_store = {}

        bot.state.ns_break_handler = None
        if break_handler:
            if isinstance(break_handler, Handler):
                bot.state.ns_collector.handlers.append(break_handler)
                bot.state.ns_break_handler = break_handler.name
            elif callable(break_handler):
                register_function_as_ns_handler(bot, break_handler)
                bot.state.ns_break_handler = get_name_from_callable(break_handler)
            else:
                bot.state.ns_collector.handlers.append(
                    bot.collector.handler_for(break_handler),
                )
                bot.state.ns_break_handler = break_handler

        if isinstance(functions, dict):
            functions_dict = functions
        else:
            functions_dict = {get_name_from_callable(func): func for func in functions}

        for name, function in functions_dict.items():
            register_function_as_ns_handler(bot, function, name)

    async def dispatch(self, message: Message, call_next: Executor) -> None:
        """Execute middleware logic.

        Arguments:
            message: incoming message.
            call_next: next executor in middleware chain.
        """
        if message.bot.state.ns_break_handler:
            break_handler = message.bot.state.ns_collector.handler_for(
                message.bot.state.ns_break_handler,
            )
            if break_handler.matches(message):
                await self.drop_next_step_handlers_chain(message)
                await break_handler(message)
                return

        try:
            next_handler, state = await self.lookup_next_handler_for_message(message)
        except (exceptions.NoMatchFound, IndexError, KeyError, RuntimeError):
            await concurrency.callable_to_coroutine(call_next, message)
            return

        key = get_chain_key_by_message(message)
        logger.bind(botx_ns_middleware=True, payload={"next_step_key": key}).info(
            "botx: found next step handler",
        )

        for state_argument in state.arguments.items():
            setattr(message.state, state_argument[0], state_argument[1])
        await next_handler(message)

    async def lookup_next_handler_for_message(
        self, message: Message,
    ) -> Tuple[Handler, NextStepHandlerState]:
        """Find handler in bot storage or in handlers.

        Arguments:
            message: message for which next step handler should be found.

        Returns:
            Found handler and state with arguments for message.
        """
        handlers: List[NextStepHandlerState] = message.bot.state.ns_store[
            get_chain_key_by_message(message)
        ]
        handler_state = handlers.pop()
        return (
            message.bot.state.ns_collector.handler_for(handler_state.name),
            handler_state,
        )

    async def drop_next_step_handlers_chain(self, message: Message) -> None:
        """Drop registered chain for message.

        Arguments:
            message: message for which chain should be dropped.
        """
        with contextlib.suppress(KeyError):
            message.bot.state.ns_store.pop(get_chain_key_by_message(message))


def get_chain_key_by_message(message: Message) -> Tuple[str, UUID, UUID, UUID]:
    """Generate key for next step handlers chain from message.

    Arguments:
        message: message from which key should be generated.

    Returns:
        Key using which handler should be found.

    Raises:
        RuntimeError: raised if key for chain can not be built.
    """
    # key is a tuple of (host, bot_id, chat_id, user_huid)
    if message.user_huid is None:
        raise RuntimeError("key for chain can be obtained only for messages from users")
    return message.host, message.bot_id, message.group_chat_id, message.user_huid


def register_function_as_ns_handler(
    bot: Bot, func: Callable, name: Optional[str] = None,
) -> None:
    """Register new function that can be called as next step handler.

    !!! warning
        This functions should not be called to dynamically register new functions in
        handlers or elsewhere, since state on different time can be changed somehow.

    Arguments:
        bot: bot that stores ns state.
        func: functions that will be called as ns handler. Will be transformed to
            coroutine if it is not already.
        name: name for new function. Will be generated from `func` if not passed.

    Raises:
        ValueError: raised if there is error to register handler.
    """
    name = name or get_name_from_callable(func)
    collector: Collector = bot.state.ns_collector
    try:
        collector.add_handler(
            body=name,
            name=name,
            handler=func,
            include_in_status=False,
            dependencies=collector.dependencies,
            dependency_overrides_provider=collector.dependency_overrides_provider,
        )
    except AssertionError as exc:
        raise ValueError(exc.args)


def register_next_step_handler(
    message: Message, func: Union[str, Callable], **ns_arguments: Any,
) -> None:
    """Register new next step handler for next message from user.

    !!! info
        While registration handler for next message this function fill first try to find
        handlers that were registered using `register_function_as_ns_handler`, then
        handlers that are registered in bot itself and then if no one was found an
        exception will be raised.

    Arguments:
        message: incoming message.
        func: function name of function which name will be retrieved to register next
            handler.
        ns_arguments: arguments that will be stored in message state while executing
            handler with next message.

    Raises:
        ValueError: raised if passed message does not include user_huid or if handler
            that should be registered as next step does not exists.
    """
    if message.user_huid is None:
        raise ValueError(
            "message for which ns handler is registered should include user_huid",
        )

    bot = message.bot
    collector: Collector = bot.state.ns_collector
    name = get_name_from_callable(func) if callable(func) else func

    try:
        collector.handler_for(name)
    except exceptions.NoMatchFound:
        raise ValueError(
            "bot does not have registered next step handler with name {0}".format(name),
        )

    key = get_chain_key_by_message(message)

    store: List[NextStepHandlerState] = bot.state.ns_store.setdefault(key, [])
    store.append(NextStepHandlerState(name=name, arguments=ns_arguments))



"""Endpoints for chats resource."""

from starlette.requests import Request
from starlette.responses import Response

from botx.clients.methods.base import APIResponse
from botx.clients.methods.v3.users.by_email import ByEmail
from botx.clients.methods.v3.users.by_huid import ByHUID
from botx.clients.methods.v3.users.by_login import ByLogin
from botx.models.users import UserFromSearch
from botx.testing.botx_mock.asgi.messages import add_request_to_collection
from botx.testing.botx_mock.asgi.responses import PydanticResponse
from botx.testing.botx_mock.binders import bind_implementation_to_method
from botx.testing.botx_mock.entities import create_test_user


@bind_implementation_to_method(ByHUID)
async def get_by_huid(request: Request) -> Response:
    """Handle retrieving information about user request.

    Arguments:
        request: HTTP request from Starlette.

    Returns:
        Response with information about user.
    """
    payload = ByHUID.parse_obj(request.query_params)
    add_request_to_collection(request, payload)
    return PydanticResponse(
        APIResponse[UserFromSearch](
            result=create_test_user(user_huid=payload.user_huid),
        ),
    )


@bind_implementation_to_method(ByEmail)
async def get_by_email(request: Request) -> Response:
    """Handle retrieving information about user request.

    Arguments:
        request: HTTP request from Starlette.

    Returns:
        Response with information about user.
    """
    payload = ByEmail.parse_obj(request.query_params)
    add_request_to_collection(request, payload)
    return PydanticResponse(
        APIResponse[UserFromSearch](result=create_test_user(email=payload.email)),
    )


@bind_implementation_to_method(ByLogin)
async def get_by_login(request: Request) -> Response:
    """Handle retrieving information about user request.

    Arguments:
        request: HTTP request from Starlette.

    Returns:
        Response with information about user.
    """
    payload = ByLogin.parse_obj(request.query_params)
    add_request_to_collection(request, payload)
    return PydanticResponse(
        APIResponse[UserFromSearch](
            result=create_test_user(ad=(payload.ad_login, payload.ad_domain)),
        ),
    )



"""Logic for extending messages and requests collections from test client."""

from molten import Settings

from botx.clients.methods.base import BotXMethod


def add_message_to_collection(settings: Settings, message: BotXMethod) -> None:
    """Add new message to messages collection.

    Arguments:
        settings: application settings with storage.
        message: message that should be added.
    """
    messages = settings["messages"]
    messages.append(message)
    add_request_to_collection(settings, message)


def add_request_to_collection(settings: Settings, api_request: BotXMethod) -> None:
    """Add new API request to requests collection.

    Arguments:
        settings: application settings with storage.
        api_request: API request that should be added.
    """
    requests = settings["requests"]
    requests.append(api_request)



"""Validators and converters for fields in builder."""

from typing import Any, BinaryIO, Optional, TextIO, Union

from botx.models import enums, events, files
from botx.models.messages.incoming_message import Sender


def validate_body_corresponds_command(body: str, values: dict) -> str:  # noqa: WPS110
    """Check that passed body can be proceed.

    Arguments:
        body: passed body.
        values: already validated validated_values.

    Returns:
        Checked passed body.
    """
    _check_system_command_properties(
        body, values.get("system_command", False), values["command_data"], values,
    )
    return body


def validate_command_type_corresponds_command(
    is_system_command: bool, values: dict,  # noqa: WPS110
) -> bool:
    """Check that command type corresponds body.

    Arguments:
        is_system_command: is command marked as system command.
        values: already validated validated_values.

    Returns:
        Checked flag.
    """
    if is_system_command:
        _check_system_command_properties(
            values["body"], is_system_command, values["command_data"], values,
        )

    return is_system_command


def convert_to_acceptable_file(
    file: Optional[Union[files.File, BinaryIO, TextIO]],
) -> Optional[files.File]:
    """Convert file to File that can be passed into message.

    Arguments:
        file: passed file.

    Returns:
        Converted file.
    """
    if isinstance(file, files.File) or file is None:
        return file

    new_file = files.File.from_file(file, filename="temp.txt")
    new_file.file_name = file.name
    return new_file


def _check_system_command_properties(
    body: str, is_system_command: bool, command_data: dict, validated_values: dict,
) -> None:
    if is_system_command:
        event = enums.SystemEvents(body)  # check that is real system event
        event_shape = events.EVENTS_SHAPE_MAP.get(event)
        if event_shape is not None:
            event_shape.parse_obj(command_data)  # check event data
        _event_checkers[event](**validated_values)  # type: ignore


def _check_common_system_event(user: Sender, **_kwargs: Any) -> None:
    error_field = ""
    if user.user_huid is not None:
        error_field = "user_huid"
    elif user.ad_login is not None:
        error_field = "ad_login"
    elif user.ad_domain is not None:
        error_field = "ad_domain"
    elif user.username is not None:
        error_field = "username"

    if error_field:
        raise ValueError(
            "user in system:chat_created can not have {0}".format(error_field),
        )


def _check_file_transfer_event(file: Optional[files.File], **_kwargs: Any) -> None:
    if file is None:
        raise ValueError("file_transfer event should have attached file")


_event_checkers = {
    enums.SystemEvents.chat_created: _check_common_system_event,
    enums.SystemEvents.added_to_chat: _check_common_system_event,
    enums.SystemEvents.file_transfer: _check_file_transfer_event,
}



"""Mixin that defines handler decorator."""

from typing import Any, Callable, Optional, Sequence

from botx.collecting.collectors.collector import Collector
from botx.dependencies.models import Depends


class HiddenHandlerMixin:
    """Mixin that defines handler decorator."""

    collector: Collector

    def hidden(  # noqa: WPS211
        self,
        handler: Optional[Callable] = None,
        *,
        command: Optional[str] = None,
        commands: Optional[Sequence[str]] = None,
        name: Optional[str] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        dependency_overrides_provider: Any = None,
    ) -> Callable:
        """Register hidden handler that won't be showed in menu.

        Arguments:
            handler: callable that will be used for executing handler.
            command: body template that will trigger this handler.
            commands: list of body templates that will trigger this handler.
            name: optional name for handler that will be used in generating body.
            dependencies: sequence of dependencies that should be executed before
                handler.
            dependency_overrides_provider: mock of callable for handler.

        Returns:
            Passed in `handler` callable.
        """
        return self.collector.hidden(
            handler=handler,
            command=command,
            commands=commands,
            name=name,
            dependencies=dependencies,
            dependency_overrides_provider=dependency_overrides_provider,
        )



"""Logic for extending messages and requests collections from test client."""

import contextlib

from starlette.requests import Request

from botx.clients.methods.base import BotXMethod


def add_message_to_collection(request: Request, message: BotXMethod) -> None:
    """Add new message to messages collection.

    Arguments:
          request: request from Starlette endpoint.
          message: message that should be added.
    """
    app = request.app
    with contextlib.suppress(AttributeError):
        app.state.messages.append(message)

    add_request_to_collection(request, message)


def add_request_to_collection(http_request: Request, api_request: BotXMethod) -> None:
    """Add new API request to requests collection.

    Arguments:
          http_request: request from Starlette endpoint.
          api_request: API request that should be added.
    """
    app = http_request.app
    with contextlib.suppress(AttributeError):
        app.state.requests.append(api_request)



"""Method for enabling stealth in chat."""
from typing import Optional
from uuid import UUID

import httpx

from botx.clients.methods.base import AuthorizedBotXMethod
from botx.clients.methods.errors import bot_is_not_admin, chat_not_found


class StealthSet(AuthorizedBotXMethod[bool]):
    """Method for enabling stealth in chat."""

    __url__ = "/api/v3/botx/chats/stealth_set"
    __method__ = "POST"
    __returning__ = bool
    __errors_handlers__ = {
        httpx.codes.FORBIDDEN: bot_is_not_admin.handle_error,
        httpx.codes.NOT_FOUND: chat_not_found.handle_error,
    }

    #: ID of chat where stealth should be enabled.
    group_chat_id: UUID

    #: should messages be shown in web.
    disable_web: bool = False

    #: time of messages burning after read.
    burn_in: Optional[int] = None

    #: time of messages burning after send.
    expire_in: Optional[int] = None



"""Endpoints for bots resource."""
from uuid import UUID

from molten import Request, Response, Settings

from botx.clients.methods.base import APIResponse
from botx.clients.methods.v2.bots.token import Token
from botx.testing.botx_mock.binders import bind_implementation_to_method
from botx.testing.botx_mock.wsgi.messages import add_request_to_collection
from botx.testing.botx_mock.wsgi.responses import PydanticResponse


@bind_implementation_to_method(Token)
def get_token(bot_id: str, request: Request, settings: Settings) -> Response:
    """Handle retrieving token from BotX API request.

    Arguments:
        bot_id: ID of bot from query params.
        request: modten request for route.
        settings: application settings with storage.

    Returns:
        Return response with new token.
    """
    signature = request.params["signature"]
    payload = Token(bot_id=UUID(bot_id), signature=signature)
    add_request_to_collection(settings, payload)
    return PydanticResponse(APIResponse[str](result="real token"))



"""Endpoints for command resource."""
from molten import RequestData, Response, Settings

from botx.clients.methods.v3.command.command_result import CommandResult
from botx.testing.botx_mock.binders import bind_implementation_to_method
from botx.testing.botx_mock.wsgi.messages import add_message_to_collection
from botx.testing.botx_mock.wsgi.responses import generate_push_response


@bind_implementation_to_method(CommandResult)
def post_command_result(request_data: RequestData, settings: Settings) -> Response:
    """Handle command result request.

    Arguments:
        request_data: parsed json data from request.
        settings: application settings with storage.

    Returns:
        Response with sync_id of pushed message.
    """
    payload = CommandResult.parse_obj(request_data)
    add_message_to_collection(settings, payload)
    return generate_push_response(payload)



"""Method for creating new chat."""

from typing import List, Optional
from uuid import UUID

import httpx

from botx.clients.methods.base import AuthorizedBotXMethod
from botx.clients.methods.errors import chat_creation_disallowed, chat_creation_error
from botx.clients.methods.extractors import extract_generated_chat_id
from botx.clients.types.response_results import ChatCreatedResult
from botx.models.enums import ChatTypes


class Create(AuthorizedBotXMethod[UUID]):
    """Method for creating new chat."""

    __url__ = "/api/v3/botx/chats/create"
    __method__ = "POST"
    __returning__ = ChatCreatedResult
    __result_extractor__ = extract_generated_chat_id
    __errors_handlers__ = {
        httpx.codes.FORBIDDEN: chat_creation_disallowed.handle_error,
        httpx.codes.UNPROCESSABLE_ENTITY: chat_creation_error.handle_error,
    }

    #: name of chat that should be created.
    name: str

    #: description of new chat.
    description: Optional[str] = None

    #: HUIDs of users that should be added into chat.
    members: List[UUID]

    #: logo image of chat.
    avatar: Optional[str] = None

    #: chat type.
    chat_type: ChatTypes



"""Definition of middleware that will generate BotX API errors depending from flag."""
from typing import Any, Callable, Tuple, Type, cast

import httpx
from molten import BaseApp, Request, Response, Route, Settings
from pydantic import BaseModel

from botx.clients.methods.base import APIErrorResponse, BotXMethod
from botx.testing.botx_mock.wsgi.responses import PydanticResponse


def _get_error_from_request(
    request: Request, app: BaseApp, settings: Settings,
) -> Tuple[Type[BotXMethod], Tuple[int, BaseModel]]:
    match = app.router.match(request.method, request.path)
    route, _ = cast(Tuple[Route, Any], match)
    method = route.handler.method  # type: ignore
    return method, settings["errors"].get(method)


def should_generate_error_response(
    request: Request, app: BaseApp, settings: Settings,
) -> bool:
    """Check if mocked API should generate error response.

    Arguments:
        request: request from molten route.
        app: molten app that serves request.
        settings: application settings with storage.

    Returns:
        Result of check.
    """
    _, status_and_error_to_raise = _get_error_from_request(request, app, settings)
    return bool(status_and_error_to_raise)


def generate_error_response(
    request: Request, app: BaseApp, settings: Settings,
) -> Response:
    """Generate error response for mocked BotX API.

    Arguments:
        request: request from molten route.
        app: molten app that serves request.
        settings: application settings with storage.

    Returns:
        Generated response.
    """
    method, response_info = _get_error_from_request(request, app, settings)
    status_code, error_data = response_info

    return PydanticResponse(
        APIErrorResponse[BaseModel](
            errors=["error from mock"], reason="asked_for_error", error_data=error_data,
        ),
        status_code="{0} {1}".format(
            status_code, httpx.codes.get_reason_phrase(status_code),
        ),
    )


def error_middleware(handler: Callable[..., Any]) -> Callable[..., Any]:
    """Middleware that will generate error response.

    Arguments:
        handler: next handler for request for molten.

    Returns:
        Created handler for request.
    """

    def decorator(request: Request, app: BaseApp, settings: Settings) -> Any:
        if should_generate_error_response(request, app, settings):
            return generate_error_response(request, app, settings)

        return handler()

    return decorator



"""Definition of messages received by bot or sent by it."""

from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import UUID

from pydantic import BaseConfig, BaseModel, Field, validator

from botx.models import events
from botx.models.enums import ChatTypes, CommandTypes, EntityTypes
from botx.models.files import File
from botx.models.forwards import Forward
from botx.models.mentions import Mention

CommandDataType = Union[
    events.ChatCreatedEvent, events.AddedToChatEvent, Dict[str, Any],
]


class Command(BaseModel):
    """Command that should be proceed by bot."""

    #: incoming text message.
    body: str

    #: was command received from user or this is system event.
    command_type: CommandTypes

    #: command payload.
    data: CommandDataType = {}  # noqa: WPS110

    #: command metadata.
    metadata: Dict[str, Any] = {}

    @property
    def command(self) -> str:
        """First word of body that was sent to bot."""
        return self.body.split(" ", 1)[0]

    @property
    def arguments(self) -> Tuple[str, ...]:
        """Words that are passed after command."""
        words = (word for word in self.body.split(" ")[1:])
        arguments = (arg for arg in words if arg and not arg.isspace())

        return tuple(arguments)

    @property
    def single_argument(self) -> str:
        """Line that passed after command."""
        body_len = len(self.command)
        return self.body[body_len:].strip()

    @property
    def data_dict(self) -> dict:
        """Command data as dictionary."""
        if isinstance(self.data, dict):
            return self.data
        return self.data.dict()


class Sender(BaseModel):
    """User that sent message to bot."""

    #: user id.
    user_huid: Optional[UUID]

    #: chat id.
    group_chat_id: UUID

    #: type of chat.
    chat_type: ChatTypes

    #: AD login of user.
    ad_login: Optional[str]

    #: AD domain of user.
    ad_domain: Optional[str]

    #: username of user.
    username: Optional[str]

    #: is user admin of chat.
    is_admin: Optional[bool]

    #: is user creator of chat.
    is_creator: Optional[bool]

    #: host from which user sent message.
    host: str

    @property
    def email(self) -> Optional[str]:
        """User email."""
        if self.ad_login and self.ad_domain:
            return "{0}@{1}".format(self.ad_login, self.ad_domain)

        return None


class Entity(BaseModel):
    """Additional entity that can be received by bot."""

    #: entity type.
    type: EntityTypes  # noqa: WPS: 125

    #: entity data.
    data: Union[Forward, Mention]  # noqa: WPS110


class IncomingMessage(BaseModel):
    """Message that was received by bot and should be handled."""

    #: message event id on which bot should answer.
    sync_id: UUID

    #: command for bot.
    command: Command

    #: file attached to message.
    file: Optional[File] = None

    #: information about user from which message was received.
    user: Sender = Field(..., alias="from")

    #: id of bot that should handle message.
    bot_id: UUID

    #: additional entities that can be received by bot.
    entities: List[Entity] = []

    class Config(BaseConfig):
        allow_population_by_field_name = True

    @validator("file", always=True, pre=True)
    def skip_file_validation(
        cls, file: Optional[Union[dict, File]],  # noqa: N805
    ) -> Optional[File]:
        """Skip validation for incoming file since users have not such limits as bot.

        Arguments:
            cls: this class.
            file: file data that should be used for building file instance.

        Returns:
            Constructed file.
        """
        if isinstance(file, File):
            return file
        elif file is not None:
            return File.construct(**file)

        return None



"""Custom extractors for responses from BotX API."""
from uuid import UUID

from botx.clients.methods.base import BotXMethod
from botx.clients.types.response_results import ChatCreatedResult, PushResult


def extract_generated_sync_id(_method: BotXMethod, push: PushResult) -> UUID:
    """Extract generated sync ID from response.

    Arguments:
        _method: method that was used for making request.
        push: push response from BotX API for generated message.

    Returns:
        Extracted sync ID.
    """
    return push.sync_id


def extract_generated_chat_id(_method: BotXMethod, response: ChatCreatedResult) -> UUID:
    """Extract generated sync ID from response.

    Arguments:
        _method: method that was used for making request.
        response: response for created chat.

    Returns:
        Extracted chat ID.
    """
    return response.chat_id



"""Middleware for retrieving tokens from BotX API before processing message."""

from botx.middlewares.base import BaseMiddleware
from botx.models.credentials import ServerCredentials
from botx.models.messages.message import Message
from botx.typing import AsyncExecutor


class AuthorizationMiddleware(BaseMiddleware):
    """Middleware for retrieving tokens from BotX API before processing message."""

    async def dispatch(self, message: Message, call_next: AsyncExecutor) -> None:
        """Obtain token for bot for handling answers to message.

        Arguments:
            message: incoming message.
            call_next: next executor in chain.
        """
        bot = message.bot
        server = bot.get_cts_by_host(message.host)
        if server.server_credentials is None:
            token = await bot.get_token(
                message.host,
                message.bot_id,
                server.calculate_signature(message.bot_id),
            )
            server.server_credentials = ServerCredentials(
                bot_id=message.bot_id, token=token,
            )
        await call_next(message)



"""Functions for inspecting signatures and parameters."""

import inspect
from typing import Any, Callable, Dict

from pydantic.typing import ForwardRef, evaluate_forwardref


def get_typed_signature(call: Callable) -> inspect.Signature:
    """Get signature for callable function with solving possible annotations.

    Arguments:
        call: callable object that will be used to get signature with annotations.

    Returns:
        Callable signature obtained.
    """
    signature = inspect.signature(call)
    global_namespace = getattr(call, "__globals__", {})
    typed_params = [
        inspect.Parameter(
            name=dependency_param.name,
            kind=dependency_param.kind,
            default=dependency_param.default,
            annotation=get_typed_annotation(dependency_param, global_namespace),
        )
        for dependency_param in signature.parameters.values()
    ]
    return inspect.Signature(typed_params)


def get_typed_annotation(
    dependency_param: inspect.Parameter, global_namespace: Dict[str, Any],
) -> Any:
    """Solve forward reference annotation for instance of `inspect.Parameter`.

    Arguments:
        dependency_param: instance of `inspect.Parameter` for which possible forward
            annotation will be evaluated.
        global_namespace: dictionary of entities that can be used for evaluating
            forward references.

    Returns:
        Parameter annotation.
    """
    annotation = dependency_param.annotation
    if isinstance(annotation, str):
        annotation = ForwardRef(annotation)
        annotation = evaluate_forwardref(annotation, global_namespace, global_namespace)
    return annotation



"""Mixin for shortcut for users resource requests."""

from typing import Optional, Tuple
from uuid import UUID

from botx.bots.mixins.requests.call_protocol import BotXMethodCallProtocol
from botx.clients.methods.v3.users.by_email import ByEmail
from botx.clients.methods.v3.users.by_huid import ByHUID
from botx.clients.methods.v3.users.by_login import ByLogin
from botx.models.messages.sending.credentials import SendingCredentials
from botx.models.users import UserFromSearch


class UsersRequestsMixin:
    """Mixin for shortcut for users resource requests."""

    async def search_user(
        self: BotXMethodCallProtocol,
        credentials: SendingCredentials,
        *,
        user_huid: Optional[UUID] = None,
        email: Optional[str] = None,
        ad: Optional[Tuple[str, str]] = None,
    ) -> UserFromSearch:
        """Search user by one of provided params for search.

        Arguments:
            credentials: credentials for making request.
            user_huid: HUID of user.
            email: email of user.
            ad: AD login and domain of user.

        Returns:
            Information about user.

        Raises:
            ValueError: raised if none  of provided params were filled.
        """
        if user_huid is not None:
            return await self.call_method(
                ByHUID(user_huid=user_huid), credentials=credentials,
            )
        elif email is not None:
            return await self.call_method(ByEmail(email=email), credentials=credentials)
        elif ad is not None:
            return await self.call_method(
                ByLogin(ad_login=ad[0], ad_domain=ad[1]), credentials=credentials,
            )

        raise ValueError("one of user_huid, email or ad query_params should be filled")



"""Definition of Starlette application that is mock for BotX API."""

from typing import Any, Dict, List, Sequence, Tuple, Type

from starlette.applications import Starlette
from starlette.middleware.base import RequestResponseEndpoint
from starlette.routing import Route

from botx.clients.methods.base import BotXMethod
from botx.testing.botx_mock.asgi.errors import ErrorMiddleware
from botx.testing.botx_mock.asgi.routes import (
    bots,
    chats,
    command,
    events,
    notification,
    users,
)
from botx.testing.typing import APIMessage, APIRequest

_ENDPOINTS: Tuple[RequestResponseEndpoint, ...] = (
    # V2
    # bots
    bots.get_token,
    # V3
    # chats
    chats.get_info,
    chats.post_add_admin_role,
    chats.post_add_user,
    chats.post_remove_user,
    chats.post_stealth_set,
    chats.post_stealth_disable,
    chats.post_create,
    # command
    command.post_command_result,
    # events
    events.post_edit_event,
    # notification
    notification.post_notification,
    notification.post_notification_direct,
    # users
    users.get_by_huid,
    users.get_by_email,
    users.get_by_login,
)


def _create_starlette_routes() -> Sequence[Route]:
    routes = []

    for endpoint in _ENDPOINTS:
        url = endpoint.method.__url__  # type: ignore  # noqa: WPS609
        method = endpoint.method.__method__  # type: ignore  # noqa: WPS609
        routes.append(Route(url, endpoint, methods=[method]))

    return routes


def get_botx_asgi_api(
    messages: List[APIMessage],
    requests: List[APIRequest],
    errors: Dict[Type[BotXMethod], Tuple[int, Any]],
) -> Starlette:
    """Generate BotX API mock.

    Arguments:
        messages: list of message that were sent from bot and should be extended.
        requests: all requests that were sent from bot.
        errors: errors to be generated by mocked API.

    Returns:
        Generated BotX API mock for using with httpx.
    """
    botx_app = Starlette(routes=list(_create_starlette_routes()))
    botx_app.add_middleware(ErrorMiddleware)
    botx_app.state.messages = messages
    botx_app.state.requests = requests
    botx_app.state.errors = errors

    return botx_app



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['botx',
 'botx.bots',
 'botx.bots.mixins',
 'botx.bots.mixins.collecting',
 'botx.bots.mixins.requests',
 'botx.clients',
 'botx.clients.clients',
 'botx.clients.interceptors',
 'botx.clients.methods',
 'botx.clients.methods.errors',
 'botx.clients.methods.v2',
 'botx.clients.methods.v2.bots',
 'botx.clients.methods.v3',
 'botx.clients.methods.v3.chats',
 'botx.clients.methods.v3.command',
 'botx.clients.methods.v3.events',
 'botx.clients.methods.v3.notification',
 'botx.clients.methods.v3.users',
 'botx.clients.types',
 'botx.collecting',
 'botx.collecting.collectors',
 'botx.collecting.collectors.mixins',
 'botx.collecting.handlers',
 'botx.dependencies',
 'botx.middlewares',
 'botx.models',
 'botx.models.messages',
 'botx.models.messages.sending',
 'botx.testing',
 'botx.testing.botx_mock',
 'botx.testing.botx_mock.asgi',
 'botx.testing.botx_mock.asgi.routes',
 'botx.testing.botx_mock.wsgi',
 'botx.testing.botx_mock.wsgi.routes',
 'botx.testing.building',
 'botx.testing.testing_client']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.16.0,<0.17.0', 'loguru>=0.5.0,<0.6.0', 'pydantic>=1.0,<2.0']

extras_require = \
{':python_version < "3.8"': ['typing-extensions>=3.7.4,<4.0.0'],
 'tests': ['starlette>=0.13.2,<0.14.0', 'molten>=1.0.1,<2.0.0']}

setup_kwargs = {
    'name': 'botx',
    'version': '0.15.10',
    'description': 'A little python framework for building bots for eXpress',
    'long_description': '<h1 align="center">pybotx</h1>\n<p align="center">\n    <em>A little python framework for building bots for eXpress messenger.</em>\n</p>\n<p align="center">\n    <a href=https://github.com/ExpressApp/pybotx>\n        <img src=https://github.com/ExpressApp/pybotx/workflows/Tests/badge.svg alt="Tests" />\n    </a>\n    <a href=https://github.com/ExpressApp/pybotx>\n        <img src=https://github.com/ExpressApp/pybotx/workflows/Styles/badge.svg alt="Styles" />\n    </a>\n    <a href="https://codecov.io/gh/ExpressApp/pybotx">\n        <img src="https://codecov.io/gh/ExpressApp/pybotx/branch/master/graph/badge.svg" alt="Coverage" />\n    </a>\n    <a href="https://github.com/ambv/black">\n        <img src="https://img.shields.io/badge/code%20style-black-000000.svg" alt="Code Style" />\n    </a>\n    <a href="https://pypi.org/project/botx/">\n        <img src="https://badge.fury.io/py/botx.svg" alt="Package version" />\n    </a>\n    <a href="https://github.com/ExpressApp/pybotx/blob/master/LICENSE">\n        <img src="https://img.shields.io/github/license/Naereen/StrapDown.js.svg" alt="License" />\n    </a>\n</p>\n\n\n---\n\n# Introduction\n\n`pybotx` is a framework for building bots for eXpress providing a mechanism for simple\nintegration with your favourite web frameworks.\n\nMain features:\n\n * Simple integration with your web apps.\n * Asynchronous API with synchronous as a fallback option.\n * 100% test coverage.\n * 100% type annotated codebase.\n\n\n**NOTE**: *This library is under active development and its API may be unstable. Please lock the version you are using at the minor update level. For example, like this in `poetry`.*\n\n```toml\n[tool.poetry.dependencies]\nbotx = "^0.15.0"\n```\n\n---\n\n## Requirements\n\nPython 3.7+\n\n`pybotx` use the following libraries:\n\n* <a href="https://github.com/samuelcolvin/pydantic" target="_blank">pydantic</a> for the data parts.\n* <a href="https://github.com/encode/httpx" target="_blank">httpx</a> for making HTTP calls to BotX API.\n* <a href="https://github.com/Delgan/loguru" target="_blank">loguru</a> for beautiful and powerful logs.\n* **Optional**. <a href="https://github.com/encode/starlette" target="_blank">Starlette</a> for tests.\n\n## Installation\n```bash\n$ pip install botx\n```\n\nOr if you are going to write tests:\n\n```bash\n$ pip install botx[tests]\n```\n\nYou will also need a web framework to create bots as the current BotX API only works with webhooks.\nThis documentation will use <a href="https://github.com/tiangolo/fastapi" target="_blank">FastAPI</a> for the examples bellow.\n```bash\n$ pip install fastapi uvicorn\n```\n\n## Example\n\nLet\'s create a simple echo bot.\n\n* Create a file `main.py` with following content:\n```python3\nfrom botx import Bot, ExpressServer, IncomingMessage, Message, Status\nfrom fastapi import FastAPI\nfrom starlette.status import HTTP_202_ACCEPTED\n\nbot = Bot(known_hosts=[ExpressServer(host="cts.example.com", secret_key="secret")])\n\n\n@bot.default(include_in_status=False)\nasync def echo_handler(message: Message) -> None:\n    await bot.answer_message(message.body, message)\n\n\napp = FastAPI()\napp.add_event_handler("shutdown", bot.shutdown)\n\n\n@app.get("/status", response_model=Status)\nasync def bot_status() -> Status:\n    return await bot.status()\n\n\n@app.post("/command", status_code=HTTP_202_ACCEPTED)\nasync def bot_command(message: IncomingMessage) -> None:\n    await bot.execute_command(message.dict())\n```\n\n* Deploy a bot on your server using uvicorn and set the url for the webhook in Express.\n```bash\n$ uvicorn main:app --host=0.0.0.0\n```\n\nThis bot will send back every your message.\n\n## License\n\nThis project is licensed under the terms of the MIT license.\n',
    'author': 'Sidnev Nikolay',
    'author_email': 'nsidnev@ccsteam.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ExpressApp/pybotx',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Definition of message object that is used in all bot handlers."""
from __future__ import annotations

from functools import partial
from typing import Any, List, Optional, Type
from uuid import UUID

from botx.bots import bots
from botx.models.datastructures import State
from botx.models.files import File
from botx.models.messages.incoming_message import (
    Command,
    Entity,
    IncomingMessage,
    Sender,
)
from botx.models.messages.sending.credentials import SendingCredentials


class _ProxyProperty:
    def __init__(self, proxy_attribute_name: str, *nested_fields: str) -> None:
        self.proxy_attribute_name = proxy_attribute_name
        self.nested_fields = list(nested_fields)

    def __get__(self, instance: Message, _owner: Type[Message]) -> Any:
        proxy_object = getattr(instance, self.proxy_attribute_name)
        accessed_result = proxy_object
        for nested_field in self.nested_fields:
            accessed_result = getattr(accessed_result, nested_field)
        return accessed_result

    def __set_name__(self, _owner: Type[Message], name: str) -> None:
        self.nested_fields.append(name)


def _proxy_property(proxy_attribute_name: str, *nested_fields: str) -> Any:
    return _ProxyProperty(proxy_attribute_name, *nested_fields)


_message_proxy_property = partial(_proxy_property, "incoming_message")
_user_proxy_property = partial(_message_proxy_property, "user")


class Message:
    """Message that is used in handlers."""

    #: incoming message from BotX.
    incoming_message: IncomingMessage

    #: bot that handles this message processing.
    bot: "bots.Bot"

    #: state of message during processing.
    state: State

    #: ID of message event.
    sync_id: UUID = _message_proxy_property()
    #: ID of bot that handles message in Express.
    bot_id: UUID = _message_proxy_property()

    #: access to command information.
    command: Command = _message_proxy_property()
    #: command body.
    body: str = _message_proxy_property("command")
    #: command data.
    data: dict = _message_proxy_property("command")  # noqa: WPS110
    #: command metadata.
    metadata: dict = _message_proxy_property("command")

    #: file from message.
    file: Optional[File] = _message_proxy_property()

    #: information about user that sent message.
    user: Sender = _message_proxy_property()
    #: HUID of user.
    user_huid: Optional[UUID] = _user_proxy_property()
    #: AD login of user.
    ad_login: Optional[str] = _user_proxy_property()
    #: AD domain of user.
    ad_domain: Optional[str] = _user_proxy_property()
    #: ID of chat from which message was received.
    group_chat_id: UUID = _user_proxy_property()
    #: type of chat.
    chat_type: str = _user_proxy_property()
    #: host of CTS from which message was received.
    host: str = _user_proxy_property()

    #: external entities in message (mentions, forwards, etc)
    entities: List[Entity] = _message_proxy_property()

    #: credentials from message for using in requests.
    credentials: SendingCredentials

    #: flag for marking that message was received from button.
    sent_from_button: bool

    def __init__(self, message: IncomingMessage, bot: "bots.Bot") -> None:
        """Initialize and update fields.

        Arguments:
            message: incoming message.
            bot: bot that handles message processing.
        """
        self.incoming_message = message
        self.bot = bot

        self.state = State()

        self.credentials = SendingCredentials(
            sync_id=self.sync_id,
            bot_id=self.bot_id,
            host=self.host,
            chat_id=self.group_chat_id,
        )

        self.sent_from_button = (
            self.data.get("ui", False) if isinstance(self.command.data, dict) else False
        )

    @classmethod
    def from_dict(cls, message: dict, bot: "bots.Bot") -> Message:
        """Parse incoming dict into message.

        Arguments:
            message: incoming message to bot as dictionary.
            bot: bot that handles message.

        Returns:
            Parsed message.
        """
        return cls(IncomingMessage(**message), bot)



"""Definition for "chat is not modifiable" error."""
from typing import NoReturn
from uuid import UUID

from httpx import Response
from pydantic import BaseModel

from botx.clients.methods.base import APIErrorResponse, BotXMethod
from botx.exceptions import BotXAPIError


class PersonalChatIsNotModifiableError(BotXAPIError):
    """Error for raising when chat is not modifiable."""

    message_template = "personal chat is not modifiable"


class PersonalChatIsNotModifiableData(BaseModel):
    """Data for error when chat is not modifiable."""

    #: ID of chat that can not be modified.
    group_chat_id: UUID


def handle_error(method: BotXMethod, response: Response) -> NoReturn:
    """Handle "chat creation error" error response.

    Arguments:
        method: method which was made before error.
        response: HTTP response from BotX API.

    Raises:
        PersonalChatIsNotModifiableError: raised always.
    """
    parsed_response = APIErrorResponse[PersonalChatIsNotModifiableData].parse_obj(
        response.json(),
    )
    error_data = parsed_response.error_data
    raise PersonalChatIsNotModifiableError(
        url=method.url,
        method=method.http_method,
        response_content=response.content,
        status_content=response.status_code,
        group_chat_id=error_data.group_chat_id,
    )



"""Logic for handling response from BotX API for real HTTP responses."""
import collections
import contextlib
from typing import TypeVar

from httpx import Response
from pydantic import ValidationError

from botx import concurrency
from botx.clients.methods.base import APIResponse, BotXMethod, ErrorHandlersInMethod

ResponseT = TypeVar("ResponseT")


def extract_result(method: BotXMethod[ResponseT], response: Response) -> ResponseT:
    """Extract result from successful response and convert it to right shape.

    Arguments:
        method: method to BotX API that was called.
        response: HTTP response from BotX API.

    Returns:
        Converted shape from BotX API.
    """
    return_shape = method.returning
    api_response = APIResponse[return_shape].parse_obj(response.json())  # type: ignore
    response_result = api_response.result
    extractor = method.result_extractor
    if extractor is not None:
        # mypy does not understand that self passed here
        return extractor(response_result)  # type: ignore

    return response_result


async def handle_error(
    method: BotXMethod, error_handlers: ErrorHandlersInMethod, response: Response,
) -> None:
    """Handle error status code from BotX API.

    Arguments:
        method: method to BotX API that was called.
        error_handlers: registered on method handlers for different responses.
        response: HTTP response from BotX API.
    """
    if not isinstance(error_handlers, collections.Sequence):
        error_handlers = [error_handlers]

    for error_handler in error_handlers:
        with contextlib.suppress(ValidationError):
            await concurrency.callable_to_coroutine(error_handler, method, response)



"""Implementation for bot classes."""

from typing import Any, Callable, Type

from botx import typing
from botx.middlewares.base import BaseMiddleware
from botx.middlewares.exceptions import ExceptionMiddleware


class MiddlewareMixin:
    """Middleware mixin for bot."""

    exception_middleware: ExceptionMiddleware

    def add_middleware(
        self, middleware_class: Type[BaseMiddleware], **kwargs: Any,
    ) -> None:
        """Register new middleware for execution before handler.

        Arguments:
            middleware_class: middleware that should be registered.
            kwargs: arguments that are required for middleware initialization.
        """
        self.exception_middleware.executor = middleware_class(
            self.exception_middleware.executor, **kwargs,
        )

    def middleware(self, handler: typing.Executor) -> Callable:
        """Register callable as middleware for request.

        Arguments:
            handler: handler for middleware logic.

        Returns:
            Passed `handler` callable.
        """
        self.add_middleware(BaseMiddleware, dispatch=handler)
        return handler



"""Custom wrapper for HTTP request for BotX API."""
from typing import Dict, Optional, Union

from pydantic.dataclasses import dataclass

PrimitiveDataType = Union[None, str, int, float, bool]


@dataclass
class HTTPRequest:
    """Wrapper for HTTP request."""

    #: HTTP method.
    method: str

    #: URL for request.
    url: str

    #: headers for request
    headers: Dict[str, str]

    #: query params for request.
    query_params: Dict[str, PrimitiveDataType]

    #: request body.
    request_data: Optional[Union[str, bytes]]

    def to_dict(self) -> dict:
        """Convert request to dictionary.

        Returns:
            Built dictionary.
        """
        return {
            "method": self.method,
            "url": self.url,
            "headers": self.headers,
            "query_params": self.query_params,
            "request_data": self.request_data,
        }



"""Definition for sync client for BotX API."""
from dataclasses import field
from typing import Any, List, Optional, TypeVar

import httpx
from httpx import Response
from loguru import logger
from pydantic.dataclasses import dataclass

from botx import concurrency
from botx.clients.clients.processing import extract_result, handle_error
from botx.clients.methods.base import BotXMethod, ErrorHandlersInMethod
from botx.converters import optional_sequence_to_list
from botx.dataclasses_config import BotXDataclassConfig
from botx.exceptions import BotXAPIError

ResponseT = TypeVar("ResponseT")


@dataclass(config=BotXDataclassConfig)
class Client:
    """Sync client for BotX API."""

    http_client: httpx.Client = field(init=False)
    interceptors: List[Any] = field(default_factory=list)

    def __post_init__(self) -> None:
        """Init or update special fields."""
        self.http_client = httpx.Client()
        self.interceptors = optional_sequence_to_list(self.interceptors)

    def call(
        self, method: BotXMethod[ResponseT], host: Optional[str] = None,
    ) -> ResponseT:
        """Make request to BotX API using passed method and return result.

        Arguments:
            method: BotX API method that should be called.
            host: override for host from method.

        Returns:
            Shape specified for method response.

        Raises:
            BotXAPIError: raised if handler for error status code was not found.
        """
        if host is not None:
            method.host = host

        response = self.execute(method)

        if httpx.codes.is_error(response.status_code):
            handlers_dict = method.error_handlers
            error_handlers = handlers_dict.get(response.status_code)
            if error_handlers is not None:
                _handle_error(method, error_handlers, response)

            raise BotXAPIError(
                url=method.url,
                method=method.http_method,
                status=response.status_code,
                response_content=response.json(),
            )

        return extract_result(method, response)

    def execute(self, method: BotXMethod) -> Response:
        """Make real HTTP request using client.

        Arguments:
            method: BotX API method that should be called.

        Returns:
            HTTP response from API.
        """
        request = method.build_http_request()
        method_name = method.__repr_name__()  # noqa: WPS609
        logger.bind(botx_client=True, payload=request.to_dict()).debug(
            "send {0} request", method_name,
        )
        return self.http_client.request(
            request.method,
            request.url,
            headers=request.headers,
            params=request.query_params,
            content=request.request_data,
        )


def _handle_error(
    method: BotXMethod, error_handlers: ErrorHandlersInMethod, response: Response,
) -> None:
    concurrency.async_to_sync(handle_error)(method, error_handlers, response)



"""Method for editing sent event."""
from uuid import UUID

from pydantic import Field

from botx.clients.methods.base import AuthorizedBotXMethod
from botx.clients.types.message_payload import UpdatePayload
from botx.clients.types.options import ResultOptions


class EditEvent(AuthorizedBotXMethod[str]):
    """Method for editing sent event."""

    __url__ = "/api/v3/botx/events/edit_event"
    __method__ = "POST"
    __returning__ = str

    #: ID of event that should be edited.
    sync_id: UUID

    #: data for editing.
    result: UpdatePayload = Field(..., alias="payload")

    #: extra options for message.
    opts: ResultOptions = ResultOptions()



"""Wrappers around param classes that are used in handlers or dependencies."""

from typing import Any, Callable

from botx.dependencies import models


def Depends(dependency: Callable, *, use_cache: bool = True) -> Any:  # noqa: N802
    """Wrap Depends param for using in handlers.

    Arguments:
        dependency: callable object that will be used in handlers or other dependencies
            instances.
        use_cache: use cache for dependency.

    Returns:
        [Depends][botx.dependencies.models.Depends] that wraps passed callable.
    """
    return models.Depends(dependency=dependency, use_cache=use_cache)



"""Definition of credentials that are used for access to BotX API."""

import base64
import hashlib
import hmac
from typing import Optional
from uuid import UUID

from pydantic import BaseModel


class ServerCredentials(BaseModel):
    """Container for credentials for bot."""

    #: bot that retrieved token from API.
    bot_id: UUID

    #: token generated for bot.
    token: str


class ExpressServer(BaseModel):
    """Server on which bot can answer."""

    #: host name of server.
    host: str

    #: secret that will be used for generating signature for bot.
    secret_key: str

    #: obtained credentials for bot.
    server_credentials: Optional[ServerCredentials] = None

    def calculate_signature(self, bot_id: UUID) -> str:
        """Calculate signature for obtaining token for bot from BotX API.

        Arguments:
            bot_id: bot for which token should be generated.

        Returns:
            Calculated signature.
        """
        return base64.b16encode(
            hmac.new(
                key=self.secret_key.encode(),
                msg=str(bot_id).encode(),
                digestmod=hashlib.sha256,
            ).digest(),
        ).decode()



"""Define several handlers for builtin exceptions from this library."""

from typing import Any

from loguru import logger

from botx.exceptions import NoMatchFound
from botx.models.messages import message as messages


async def dependency_failure_exception_handler(*_: Any) -> None:
    """Just do nothing if there is this error, since it's just a signal for stop.

    Arguments:
        _: default arguments passed to exception handler.
    """


async def no_match_found_exception_handler(
    exception: NoMatchFound, message: messages.Message,
) -> None:
    """Log that handler was not found.

    Arguments:
        exception: raised NoMatchFound exception.
        message: message on which processing error was raised.
    """
    logger.info("handler for {0!r} was not found", exception.search_param)



"""Builder for messages in tests."""

import uuid
from dataclasses import field
from typing import Any, List, Optional

from pydantic import BaseConfig, validator
from pydantic.dataclasses import dataclass

from botx import ChatMention, Mention, MentionTypes, UserMention
from botx.models import enums
from botx.models.enums import ChatTypes, EntityTypes
from botx.models.messages.incoming_message import (
    Command,
    Entity,
    IncomingMessage,
    Sender,
)
from botx.testing.building.validators import (
    convert_to_acceptable_file,
    validate_body_corresponds_command,
    validate_command_type_corresponds_command,
)


def _build_default_user() -> Sender:
    return Sender(
        user_huid=uuid.uuid4(),
        group_chat_id=uuid.uuid4(),
        chat_type=ChatTypes.chat,
        ad_login="test_user",
        ad_domain="example.com",
        username="Test User",
        is_admin=True,
        is_creator=True,
        host="cts.example.com",
    )


class BuilderConfig(BaseConfig):
    """Config for builder dataclass."""

    validate_assignment = True


@dataclass(config=BuilderConfig)
class MessageBuilder:
    """Builder for command message for bot."""

    bot_id: uuid.UUID = field(default_factory=uuid.uuid4)

    command_data: dict = field(default_factory=dict)
    system_command: bool = field(default=False)
    file: Optional[Any] = field(default=None)
    user: Sender = field(default_factory=_build_default_user)
    entities: List[Entity] = field(default_factory=list)
    body: str = field(default="")

    _body_and_command_validator = validator("body", always=True)(
        validate_body_corresponds_command,
    )
    _command_type_and_data_validator = validator("system_command", always=True)(
        validate_command_type_corresponds_command,
    )
    _file_converter = validator("file", always=True)(convert_to_acceptable_file)

    @property
    def message(self) -> IncomingMessage:
        """Message that was built by builder."""
        command_type = (
            enums.CommandTypes.system
            if self.system_command
            else enums.CommandTypes.user
        )
        command = Command(
            body=self.body, command_type=command_type, data=self.command_data,
        )
        return IncomingMessage(
            sync_id=uuid.uuid4(),
            command=command,
            file=self.file,
            bot_id=self.bot_id,
            user=self.user,
            entities=self.entities,
        )

    def mention_contact(self, user_huid: uuid.UUID) -> None:
        """Add contact mention to message for bot.

        Arguments:
            user_huid: huid of user to mention.
        """
        self.entities.append(
            Entity(
                type=EntityTypes.mention,
                data=Mention(
                    mention_data=UserMention(user_huid=user_huid),
                    mention_type=MentionTypes.contact,
                ),
            ),
        )

    def mention_user(self, user_huid: uuid.UUID) -> None:
        """Add user mention to message for bot.

        Arguments:
            user_huid: huid of user to mention.
        """
        self.entities.append(
            Entity(
                type=EntityTypes.mention,
                data=Mention(mention_data=UserMention(user_huid=user_huid)),
            ),
        )

    def mention_chat(self, chat_id: uuid.UUID) -> None:
        """Add chat mention to message for bot.

        Arguments:
            chat_id: id of chat to mention.
        """
        self.entities.append(
            Entity(
                type=EntityTypes.mention,
                data=Mention(
                    mention_data=ChatMention(group_chat_id=chat_id),
                    mention_type=MentionTypes.chat,
                ),
            ),
        )



"""Definition for mixin with system events decorator."""
from typing import Any, Callable, Optional, Sequence, cast

from botx.collecting.collectors.mixins.handler import HandlerDecoratorProtocol
from botx.dependencies.models import Depends
from botx.models.enums import SystemEvents

try:
    from typing import Protocol  # noqa: WPS433
except ImportError:
    from typing_extensions import (  # type: ignore  # noqa: WPS433, WPS440, F401
        Protocol,
    )


class SystemEventsHandlerMixin:
    """Mixin that defines system events handler decorator."""

    def system_event(  # noqa: WPS211
        self,
        handler: Optional[Callable] = None,
        *,
        event: Optional[SystemEvents] = None,
        events: Optional[Sequence[SystemEvents]] = None,
        name: Optional[str] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        dependency_overrides_provider: Any = None,
    ) -> Callable:
        """Register handler for system event.

        Arguments:
            handler: callable that will be used for executing handler.
            event: event for triggering this handler.
            events: a sequence of events that will trigger handler.
            name: optional name for handler that will be used in generating body.
            dependencies: sequence of dependencies that should be executed before
                handler.
            dependency_overrides_provider: mock of callable for handler.

        Returns:
            Passed in `handler` callable.

        Raises:
            AssertionError: raised if nor event or events passed.
        """
        if not (event or events):
            raise AssertionError("at least one event should be passed")

        return cast(HandlerDecoratorProtocol, self).handler(
            handler=handler,
            command=event.value if event else None,
            commands=[event.value for event in events] if events else None,
            name=name,
            include_in_status=False,
            dependencies=dependencies,
            dependency_overrides_provider=dependency_overrides_provider,
        )

    def chat_created(
        self,
        handler: Optional[Callable] = None,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
        dependency_overrides_provider: Any = None,
    ) -> Callable:
        """Register handler for `system:chat_created` event.

        Arguments:
            handler: callable that will be used for executing handler.
            dependencies: sequence of dependencies that should be executed before
                handler.
            dependency_overrides_provider: mock of callable for handler.

        Returns:
            Passed in `handler` callable.
        """
        return self.system_event(
            handler=handler,
            event=SystemEvents.chat_created,
            name=SystemEvents.chat_created.value,
            dependencies=dependencies,
            dependency_overrides_provider=dependency_overrides_provider,
        )

    def file_transfer(
        self,
        handler: Optional[Callable] = None,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
        dependency_overrides_provider: Any = None,
    ) -> Callable:
        """Register handler for `file_transfer` event.

        Arguments:
            handler: callable that will be used for executing handler.
            dependencies: sequence of dependencies that should be executed before
                handler.
            dependency_overrides_provider: mock of callable for handler.

        Returns:
            Passed in `handler` callable.
        """
        return self.system_event(
            handler=handler,
            event=SystemEvents.file_transfer,
            name=SystemEvents.file_transfer.value,
            dependencies=dependencies,
            dependency_overrides_provider=dependency_overrides_provider,
        )

    def added_to_chat(
        self,
        handler: Optional[Callable] = None,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
        dependency_overrides_provider: Any = None,
    ) -> Callable:
        """Register handler for `added_to_chat` event.

        Arguments:
            handler: callable that will be used for executing handler.
            dependencies: sequence of dependencies that should be executed before
                handler.
            dependency_overrides_provider: mock of callable for handler.

        Returns:
            Passed in `handler` callable.
        """
        return self.system_event(
            handler=handler,
            event=SystemEvents.added_to_chat,
            name=SystemEvents.added_to_chat.value,
            dependencies=dependencies,
            dependency_overrides_provider=dependency_overrides_provider,
        )



"""Validators and extractors for Handler fields."""
import inspect
from typing import Any, Callable, List, Optional

from botx.collecting.handlers.name_generators import get_name_from_callable
from botx.dependencies.models import Dependant, Depends, get_dependant
from botx.dependencies.solving import get_executor


def validate_body_for_status(executor: Callable, values: dict) -> Callable:
    """Validate that body is acceptable for status.

    Arguments:
        executor: executor that will be just returned from validator.
        values: already checked validated_values.

    Returns:
        Passed executor.

    Raises:
        ValueError: raised if body is not acceptable for status.
    """
    include_in_status = values["include_in_status"]
    if not include_in_status:
        return executor

    body = values["body"]

    if not body.startswith("/"):
        raise ValueError("public commands should start with leading slash")

    slash_part_of_body = body[: -len(body.strip("/"))]
    if slash_part_of_body.count("/") != 1:
        raise ValueError("command body can contain only single leading slash")

    if len(body.split()) != 1:
        raise ValueError("public commands should contain only one word")

    return executor


def retrieve_name_for_handler(name: Optional[str], handler: Callable) -> str:
    """Retrieve name for handler.

    Arguments:
        name: passed name for handler.
        handler: handler that will be used to generate name.

    Returns:
        Name for handler.
    """
    return name or get_name_from_callable(handler)


def retrieve_full_description_for_handler(
    full_description: Optional[str], handler: Callable,
) -> str:
    """Retrieve full description for handler.

    Arguments:
        full_description: passed name for handler.
        handler: handler from which docstring will be used.

    Returns:
        Full description for handler.
    """
    return full_description or inspect.cleandoc(handler.__doc__ or "")


def check_handler_is_function(handler: Callable) -> Callable:
    """Check handler can be proceed by library.

    Library can handle functions and methods as handlers.

    Arguments:
        handler: passed handler callable.

    Returns:
        Passed handler.

    Raises:
        ValueError: raised if handler is not acceptable.
    """
    if not (inspect.isfunction(handler) or inspect.ismethod(handler)):
        raise ValueError("handler must be a function or method")

    return handler


def retrieve_dependant(handler: Callable, dependencies: List[Depends]) -> Dependant:
    """Retrieve dependant for handler.

    Arguments:
        handler: handler for which dependant should be created.
        dependencies: passed background dependencies.

    Returns:
        Generated dependant object.
    """
    dependant = get_dependant(call=handler)
    for index, depends in enumerate(dependencies):
        dependant.dependencies.insert(
            index, get_dependant(call=depends.dependency, use_cache=depends.use_cache),
        )

    return dependant


def retrieve_executor(
    dependant: Dependant, dependency_overrides_provider: Any,
) -> Callable:
    """Retrieve executor for handler.

    Arguments:
        dependant: dependant that will be used to generate executor.
        dependency_overrides_provider: overrider for dependencies.

    Returns:
        Generated executor.
    """
    return get_executor(
        dependant=dependant,
        dependency_overrides_provider=dependency_overrides_provider,
    )



"""Definition of entities for using in tests."""

from botx.testing.building.builder import MessageBuilder

try:
    from botx.testing.testing_client.client import TestClient  # noqa: WPS433
except ImportError:
    TestClient = None  # type: ignore  # noqa: WPS440

__all__ = ("TestClient", "MessageBuilder")  # noqa: WPS410



"""Definition of client for testing."""
from typing import Tuple, Union

from botx.clients.methods.v3.command.command_result import CommandResult
from botx.clients.methods.v3.events.edit_event import EditEvent
from botx.clients.methods.v3.notification.direct_notification import NotificationDirect
from botx.clients.methods.v3.notification.notification import Notification
from botx.testing.testing_client.base import BaseTestClient
from botx.testing.typing import APIMessage, APIRequest


class TestClient(BaseTestClient):
    """Test client for testing bots."""

    # https://docs.pytest.org/en/latest/changelog.html#changes
    # Allow to skip test classes from being collected
    __test__: bool = False

    @property
    def requests(self) -> Tuple[APIRequest, ...]:
        """Return all requests that were sent by bot.

        Returns:
            Sequence of requests that were sent from bot.
        """
        return tuple(request.copy(deep=True) for request in self._requests)

    @property
    def messages(self) -> Tuple[APIMessage, ...]:
        """Return all entities that were sent by bot.

        Returns:
            Sequence of messages that were sent from bot.
        """
        return tuple(message.copy(deep=True) for message in self._messages)

    @property
    def command_results(self) -> Tuple[CommandResult, ...]:
        """Return all command results that were sent by bot.

        Returns:
            Sequence of command results that were sent from bot.
        """
        return tuple(
            message for message in self.messages if isinstance(message, CommandResult)
        )

    @property
    def notifications(self) -> Tuple[Union[Notification, NotificationDirect], ...]:
        """Return all notifications that were sent by bot.

        Returns:
            Sequence of notifications that were sent by bot.
        """
        return tuple(
            message
            for message in self.messages
            if isinstance(message, (Notification, NotificationDirect))
        )

    @property
    def message_updates(self) -> Tuple[EditEvent, ...]:
        """Return all updates that were sent by bot.

        Returns:
            Sequence of updates that were sent by bot.
        """
        return tuple(
            message for message in self.messages if isinstance(message, EditEvent)
        )



"""Definition of enums that are used across different components of this library."""

from enum import Enum


class SystemEvents(Enum):
    """System enums that bot can retrieve from BotX API in message.

    !!! info
        NOTE: `file_transfer` is not a system event, but it is logical to place it in
        this enum.
    """

    #: `system:chat_created` event.
    chat_created = "system:chat_created"

    #: `system:added_to_chat` event.
    added_to_chat = "system:added_to_chat"

    #: `file_transfer` message.
    file_transfer = "file_transfer"


class CommandTypes(str, Enum):
    """Enum that specify from whom command was received."""

    #: command received from user.
    user = "user"

    #: command received from system.
    system = "system"


class ChatTypes(str, Enum):
    """Enum for type of chat."""

    #: private chat for user with bot.
    chat = "chat"

    #: chat with several users.
    group_chat = "group_chat"

    #: channel chat.
    channel = "channel"


class UserKinds(str, Enum):
    """Enum for type of user."""

    #: normal user.
    user = "user"

    #: normal user, but will present if all users in chat are from the same CTS.
    cts_user = "cts_user"

    #: bot user.
    bot = "botx"


class Statuses(str, Enum):
    """Enum for status of operation in BotX API."""

    #: operation was successfully proceed.
    ok = "ok"

    #: there was an error while processing operation.
    error = "error"


class EntityTypes(str, Enum):
    """Types for entities that could be received by bot."""

    #: mention entity.
    mention = "mention"

    #: forward entity.
    forward = "forward"


class MentionTypes(str, Enum):
    """Enum for available validated_values in mentions."""

    #: mention single user from chat in message.
    user = "user"

    #: mention user by user_huid.
    contact = "contact"

    #: mention chat in message.
    chat = "chat"

    #: mention channel in message.
    channel = "channel"



"""Forward entities that can be received in message."""

from datetime import datetime
from typing import Optional
from uuid import UUID

from pydantic import BaseModel

from botx.models.enums import ChatTypes


class Forward(BaseModel):
    """Forward in message."""

    #: ID of chat from which forward received.
    group_chat_id: UUID

    #: ID of user that is author of message.
    sender_huid: UUID

    #: type of forward.
    forward_type: ChatTypes

    #: name of original chat.
    source_chat_name: Optional[str]

    #: id of original message event.
    source_sync_id: Optional[UUID]

    #: id of event creation.
    source_inserted_at: datetime



"""Generators for names for handler."""

import inspect
import re
from typing import Callable


def get_body_from_name(name: str) -> str:
    """Get auto body from given handler name in format `/word-word`.

    Examples:
        ```
        >>> get_body_from_name("HandlerFunction")
        "handler-function"
        >>> get_body_from_name("handlerFunction")
        "handler-function"
        ```
    Arguments:
        name: name of handler for which body should be generated.
    """
    splited_words = re.findall(r"^[a-z\d_\-]+|[A-Z\d_\-][^A-Z\d_\-]*", name)
    joined_body = "-".join(splited_words)
    dashed_body = joined_body.replace("_", "-")
    return "/{0}".format(re.sub(r"-+", "-", dashed_body).lower())


def get_name_from_callable(handler: Callable) -> str:
    """Get auto name from given callable object.

    Arguments:
        handler: callable object that will be used to retrieve auto name for handler.

    Returns:
        Name obtained from callable.
    """
    is_function = inspect.isfunction(handler)
    is_method = inspect.ismethod(handler)
    is_class = inspect.isclass(handler)
    if is_function or is_method or is_class:
        return handler.__name__
    return handler.__class__.__name__



"""Protocol for using in mixins to make mypy and typing happy."""
from typing import Any, Optional

from botx.clients.methods.base import BotXMethod
from botx.models.messages.sending.credentials import SendingCredentials

try:
    from typing import Protocol  # noqa: WPS433
except ImportError:
    from typing_extensions import (  # type: ignore  # noqa: WPS433, WPS440, F401
        Protocol,
    )


class BotXMethodCallProtocol(Protocol):
    """Protocol for using in mixins to make mypy and typing happy."""

    async def call_method(
        self,
        method: BotXMethod[Any],
        *,
        host: Optional[str] = None,
        token: Optional[str] = None,
        credentials: Optional[SendingCredentials] = None,
    ) -> Any:
        """Send request to BotX API through bot's async client."""
